@@ -1,5 +1,5 @@
 package Date::Manip::Recur;
-# Copyright (c) 1998-2015 Sullivan Beck. All rights reserved.
+# Copyright (c) 1998-2019 Sullivan Beck. All rights reserved.
 # This program is free software; you can redistribute it and/or modify
 # it under the same terms as Perl itself.
 
@@ -26,7 +26,7 @@ use Date::Manip::Base;
 use Date::Manip::TZ;
 
 our $VERSION;
-$VERSION='6.52';
+$VERSION='6.78';
 END { undef $VERSION; }
 
 ########################################################################
@@ -349,7 +349,6 @@ sub frequency {
          return 1;
       }
 
-      my($l,$r);
       my $err = $self->_parse_lang($string);
       if ($err) {
          $$self{'err'} = "[frequency] Invalid frequency string";
@@ -362,6 +361,10 @@ sub frequency {
    # to 1.
 
    if (@int) {
+      for my $i (@int) {
+         $i += 0;
+      }
+
       TEST_INT: {
          for my $i (@int) {
             last TEST_INT if ($i);
@@ -370,8 +373,8 @@ sub frequency {
       }
    }
 
-   # If @int contains 2 or 3 elements, move a trailing 0 to the start
-   # of @rtime.
+   # If @int contains 2 or 3 elements and ends in 0, move the trailing
+   # 0 to the start of @rtime.
    #
    #   Y:M:0 * D:H:MN:S  =>  Y:M * 0:D:H:MN:S
 
@@ -382,6 +385,75 @@ sub frequency {
       unshift(@rtime,0);
    }
 
+   # We need to know what the valid values of M, W, and D are.
+   #
+   # Month can be:
+   #    moy     : 1 to 12                  (month of the year)
+   #
+   # Week can be:
+   #    woy     : 1 to 53  or  -1 to -53   (week of the year)
+   #    wom     : 1 to 5   or  -1 to -5    (week of the month)
+   #
+   # Day can be:
+   #    doy     : 1 to 366 or  -1 to -366  (day of the year)
+   #    dom     : 1 to DiM or  -1 to -31   (day of the month)
+   #    dow     : 1 to 7                   (day of the week)
+   #
+   # Other values must be zero or positive.
+
+   my @ftype = ('y','m','w','d','h','mn','s');
+   my @vtype = ('' ,'' ,'' ,'' ,'' ,''  ,'');
+
+   my ($y,$m,$w,$d,$h,$mn,$s) = (@int,@rtime);
+
+   if (@rtime == 7) {
+      $vtype[0] = 'y';
+   }
+
+   if (@rtime >= 6) {
+      if ($m) {
+         $vtype[1] = 'moy';
+      } else {
+         $vtype[1] = 'zero';
+      }
+   }
+
+   if (@rtime >= 5) {
+      if ($w) {
+         if ($m) {
+            $vtype[2] = 'wom';
+         } else {
+            $vtype[2] = 'woy';
+         }
+      } else {
+         $vtype[2] = 'zero';
+      }
+   }
+
+   if (@rtime >= 4) {
+      if ($d) {
+         if ($w) {
+            $vtype[3] = 'dow';
+         } elsif ($m) {
+            $vtype[3] = 'dom';
+         } else {
+            $vtype[3] = 'doy';
+         }
+      } else {
+         $vtype[3] = 'zero';
+      }
+   }
+
+   if (@rtime >= 3) {
+      $vtype[4] = 'time';
+   }
+   if (@rtime >= 2) {
+      $vtype[5] = 'time';
+   }
+   if (@rtime) {
+      $vtype[6] = 'time';
+   }
+
    # Test the format of @rtime.
    #
    # Turn it to:
@@ -392,31 +464,37 @@ sub frequency {
 
    my $rfieldrx = $self->_rx('rfield');
    my $rrangerx = $self->_rx('rrange');
-   my @type     = qw(y m w d h mn s);
-   while ($#type > $#rtime) {
-      shift(@type);
-   }
 
-   foreach my $rfield (@rtime) {
-      my $type = shift(@type);
+   my $i        = -1;
+   foreach my $f (@int,@rtime) {
+      $i++;
+      my $vtype = $vtype[$i];
+      my $type  = $ftype[$i];
 
-      if ($rfield !~ $rfieldrx) {
+      # $f          3   -6   2-3   1,5-6
+      # $type       y m w d h mn s
+      # $vtype      '' dom woy time      ('' is a frequency field)
+
+      # Ignore the frequency part
+      next  if (! $vtype);
+
+      if ($f  &&  $f !~ $rfieldrx) {
          $$self{'err'} = "[frequency] Invalid rtime string";
          return 1;
       }
 
-      my @rfield = split(/,/,$rfield);
+      my @rfield = split(/,/,$f);
       my @val;
 
       foreach my $vals (@rfield) {
          if ($vals =~ $rrangerx) {
-            my ($num1,$num2) = ($1,$2);
+            my ($num1,$num2) = ($1+0,$2+0);
 
-            if ( ($num1 < 0  ||  $num2 < 0)  &&
-                 ($type ne 'w'  &&  $type ne 'd') ) {
-               $$self{'err'} = "[frequency] Negative values allowed for day/week";
-               return 1;
-            }
+            my $err = $self->_frequency_values($num1,$type,$vtype);
+            return $err  if ($err);
+
+            $err    = $self->_frequency_values($num2,$type,$vtype);
+            return $err  if ($err);
 
             if ( ($num1 > 0  &&  $num2 > 0)  ||
                  ($num1 < 0  &&  $num2 < 0) ) {
@@ -430,16 +508,16 @@ sub frequency {
             }
 
          } else {
-            if ($vals < 0  &&
-                 ($type ne 'w'  &&  $type ne 'd') ) {
-               $$self{'err'} = "[frequency] Negative values allowed for day/week";
-               return 1;
-            }
+            $vals += 0;
+
+            my $err = $self->_frequency_values($vals,$type,$vtype);
+            return $err  if ($err);
+
             push(@val,$vals);
          }
       }
 
-      $rfield = [ @val ];
+      $f = [ @val ];
    }
 
    # Store it
@@ -496,6 +574,81 @@ sub frequency {
    return 0;
 }
 
+sub _frequency_values {
+   my($self,$num,$type,$vtype) = @_;
+   my $err;
+
+   if ($type eq 'y') {
+      if ($vtype eq 'y') {
+         if ($num < 0  ||  $num > 9999) {
+            $$self{'err'} = "[frequency] Year must be in the range 1-9999";
+            return 1;
+         }
+      }
+
+   } elsif ($type eq 'm') {
+      if ($vtype eq 'moy') {
+         if ($num < 1  ||  $num > 12) {
+            $$self{'err'} = "[frequency] Month of year must be 1-12";
+            return 1;
+         }
+      }
+
+   } elsif ($type eq 'w') {
+      if ($vtype eq 'woy') {
+         if ($num == 0  ||  $num > 53  ||  $num < -53) {
+            $$self{'err'} = "[frequency] Week of year must be 1-53 or -1 to -53";
+            return 1;
+         }
+
+      } elsif ($vtype eq 'wom') {
+         if ($num == 0  ||  $num > 5  ||  $num < -5) {
+            $$self{'err'} = "[frequency] Week of month must be 1-5 or -1 to -5";
+            return 1;
+         }
+
+      }
+
+   } elsif ($type eq 'd') {
+      if ($vtype eq 'dow') {
+         if ($num < 1  ||  $num > 7) {
+            $$self{'err'} = "[frequency] Day of week must be 1-7";
+            return 1;
+         }
+
+      } elsif ($vtype eq 'dom') {
+         if ($num == 0  ||  $num > 31  ||  $num < -31) {
+            $$self{'err'} = "[frequency] Day of month must be 1-31 or -1 to -31";
+            return 1;
+         }
+
+      } elsif ($vtype eq 'doy') {
+         if ($num == 0  ||  $num > 366  ||  $num < -366) {
+            $$self{'err'} = "[frequency] Day of year must be 1-366 or -1 to -366";
+            return 1;
+         }
+      }
+
+   } elsif ($type eq 'h') {
+      if ($vtype eq 'time') {
+         if ($num < 0  ||  $num > 23) {
+            $$self{'err'} = "[frequency] Hour must be 0-23";
+            return 1;
+         }
+      }
+
+   } else {
+      if ($vtype eq 'time') {
+         if ($num < 0  ||  $num > 59) {
+            $$self{'err'} = "[frequency] Minute/second must be 0-59";
+            return 1;
+         }
+      }
+   }
+
+   return 0;
+}
+
 sub _parse_lang {
    my($self,$string) = @_;
    my $dmt           = $$self{'tz'};
@@ -763,6 +916,7 @@ sub next {
              defined $$self{'data'}{'end'}) {
 
             my $n = $self->_locate_n('first');
+            return (undef,'Not found')  if ($$self{'err'});
             $$self{'data'}{'curr'} = $n-1;
 
          } else {
@@ -815,6 +969,7 @@ sub prev {
              defined $$self{'data'}{'end'}) {
 
             my $n = $self->_locate_n('last');
+            return (undef,'Not found')  if ($$self{'err'});
             $$self{'data'}{'curr'} = $n+1;
 
          } else {
@@ -887,7 +1042,9 @@ sub dates {
    my($end,$first,$last,@dates);
 
    $first = $self->_locate_n('first');
+   return ()  if ($$self{'err'});
    $last  = $self->_locate_n('last');
+   return ()  if ($$self{'err'});
 
    if (defined($first)  &&  defined($last)) {
       for (my $n = $first; $n <= $last; $n++) {
@@ -1125,7 +1282,7 @@ sub _rx {
             $rx eq 'rnum'   ||
             $rx eq 'rrange') {
 
-      my $num    = '\-?\d+';
+      my $num    = '[+-]?\d+';
       my $range  = "$num\-$num";
       my $val    = "(?:$range|$num)";
       my $vals   = "$val(?:,$val)*";
@@ -1346,10 +1503,10 @@ sub _apply_rtime_mods {
 
          #  *1:2:3:4       in Feb 0001 on the 3rd Thur of the month
          #  *0:2:3:4       on the 3rd Thur of Feb in the curr year
-         #  *1:2:3:0       the 3rd occurence of FirstDay in Feb 0001
-         #  *0:2:3:0       the 3rd occurence of FirstDay in Feb of curr year
+         #  *1:2:3:0       the 3rd occurrence of FirstDay in Feb 0001
+         #  *0:2:3:0       the 3rd occurrence of FirstDay in Feb of curr year
          #   1*2:3:4       every year in Feb on the 3rd Thur
-         #   1*2:3:0       every year on the 3rd occurence of FirstDay in Feb
+         #   1*2:3:0       every year on the 3rd occurrence of FirstDay in Feb
 
          ($err,@m)  = $self->_rtime_values('m',$mf);
          return ()  if ($err);
@@ -1400,9 +1557,9 @@ sub _apply_rtime_mods {
 
       } else {
 
-         #   0:2*3:0       every 2 months on the 3rd occurence of FirstDay
+         #   0:2*3:0       every 2 months on the 3rd occurrence of FirstDay
          #   0:2*3:4       every 2 months on the 3rd Thur of the month
-         #   1:2*3:0       every 1 year, 2 months on 3rd occurence of FirstDay
+         #   1:2*3:0       every 1 year, 2 months on 3rd occurrence of FirstDay
          #   1:2*3:4       every 1 year, 2 months on the 3rd Thur of the month
 
          if ($d_empty) {
@@ -1800,12 +1957,66 @@ sub _nth_interval {
 sub _locate_n {
    my($self,$op) = @_;
 
-   return $$self{'data'}{$op}  if (defined $$self{'data'}{$op}  ||
-                                   $$self{'data'}{'noint'} == 2);
+   return $$self{'data'}{$op}  if (defined $$self{'data'}{$op});
+
+   my $start  = $$self{'data'}{'start'};
+   my $end    = $$self{'data'}{'end'};
+   my $unmod  = $$self{'data'}{'unmod_range'};
+   my $dmt    = $$self{'tz'};
+   my $dmb    = $$dmt{'base'};
+   my $maxatt = $dmb->_config('maxrecurattempts');
+
+   if ($$self{'data'}{'noint'} == 2) {
+      # If there is no interval, then we have calculated all the dates
+      # possible.  Work with them only.
+
+      my($i,$first,$last);
+
+      # Find the first date in the interval
+
+      $i = 0;
+      while (1) {
+         last  if (! exists $$self{'data'}{'dates'}{$i});
+         my $date = $$self{'data'}{'dates'}{$i};
+         if ($date->cmp($start) == -1) {
+            # date < start   : move to the next one
+            $i++;
+            next;
+         } elsif ($date->cmp($end) == 1) {
+            # date > end     : we're done
+            last;
+         } else {
+            # start <= date <= end    : this is the first one
+            $first = $i;
+            last;
+         }
+      }
+
+      # If we found one, find the last one
+
+      if (defined($first)) {
+         $i       = $first;
+         $last    = $i;
+         while (1) {
+            last  if (! exists $$self{'data'}{'dates'}{$i});
+            my $date = $$self{'data'}{'dates'}{$i};
+            if ($date->cmp($end) == 1) {
+               # date > end    : we're done
+               last;
+            } else {
+               # date <= end   : this might be the last one
+               $last = $i;
+               $i++;
+               next;
+            }
+         }
+      }
+
+      $$self{'data'}{'first'} = $first;
+      $$self{'data'}{'last'}  = $last;
+      return $$self{'data'}{$op}
+   }
 
-   my $start = $$self{'data'}{'start'};
-   my $end   = $$self{'data'}{'end'};
-   my $unmod = $$self{'data'}{'unmod_range'};
 
    # Given interval date Idate(n) produces event dates: Date(f)..Date(l)
    #
@@ -1841,20 +2052,44 @@ sub _locate_n {
 
       $first_int    = 0;
       if ($unmod) {
+         my $n = 0;
          while (1) {
+            if ($n > $maxatt) {
+               $$self{'err'} =
+                 "[_locate_n] Unable to find an interval in $maxatt attempts";
+               return;
+            }
             $self->_nth_interval($first_int);
             my $date = $$self{'data'}{'idate'}{$first_int}[0];
-            last  if (defined $date  &&  $date->cmp($start) < 0);
+            if (defined($date)) {
+               $n = 0;
+               last  if ($date->cmp($start) < 0);
+            } else {
+               $n++;
+            }
             $first_int--;
          }
 
       } else {
+         my $n = 0;
          while (1) {
+            if ($n > $maxatt) {
+               $$self{'err'} =
+                 "[_locate_n] Unable to find an interval in $maxatt attempts";
+               return;
+            }
             $self->_nth_interval($first_int);
             my $ptr   = $$self{'data'}{'idate'}{$first_int}[2];
             if (defined $ptr) {
                my $date  = $$self{'data'}{'dates'}{$ptr};
-               last  if (defined $date  &&  $date->cmp($start) < 0);
+               if (defined($date)) {
+                  $n = 0;
+                  last  if ($date->cmp($start) < 0);
+               } else {
+                  $n++;
+               }
+            } else {
+               $n++;
             }
             $first_int--;
          }
@@ -1864,21 +2099,45 @@ sub _locate_n {
       #   i.e. Date(y) >= start  for modified dates
 
       if ($unmod) {
+         my $n = 0;
          while (1) {
+            if ($n > $maxatt) {
+               $$self{'err'} =
+                 "[_locate_n] Unable to find an interval in $maxatt attempts";
+               return;
+            }
             $self->_nth_interval($first_int);
             my $date = $$self{'data'}{'idate'}{$first_int}[0];
-            last  if (defined $date  &&  $date->cmp($start) >= 0);
+            if (defined($date)) {
+               $n = 0;
+               last  if ($date->cmp($start) >= 0);
+            } else {
+               $n++;
+            }
             $first_int++;
          }
          $first = $$self{'data'}{'idate'}{$first_int}[1];
 
       } else {
+         my $n = 0;
          while (1) {
+            if ($n > $maxatt) {
+               $$self{'err'} =
+                 "[_locate_n] Unable to find an interval in $maxatt attempts";
+               return;
+            }
             $self->_nth_interval($first_int);
             my $ptr   = $$self{'data'}{'idate'}{$first_int}[2];
             if (defined $ptr) {
                my $date  = $$self{'data'}{'dates'}{$ptr};
-               last  if (defined $date  &&  $date->cmp($start) >= 0);
+               if (defined($date)) {
+                  $n = 0;
+                  last  if ($date->cmp($start) >= 0);
+               } else {
+                  $n++;
+               }
+            } else {
+               $n++;
             }
             $first_int++;
          }
@@ -1899,10 +2158,21 @@ sub _locate_n {
       $last_int = $first_int;
 
       if ($unmod) {
+         my $n = 0;
          while (1) {
+            if ($n > $maxatt) {
+               $$self{'err'} =
+                 "[_locate_n] Unable to find an interval in $maxatt attempts";
+               return;
+            }
             $self->_nth_interval($last_int);
             my $date = $$self{'data'}{'idate'}{$last_int}[0];
-            last  if (defined $date  &&  $date->cmp($end) > 0);
+            if (defined($date)) {
+               $n = 0;
+               last  if ($date->cmp($end) > 0);
+            } else {
+               $n++;
+            }
             $last_int++;
          }
          $last_int--;
@@ -1917,12 +2187,25 @@ sub _locate_n {
          }
 
       } else {
+         my $n = 0;
          while (1) {
+            if ($n > $maxatt) {
+               $$self{'err'} =
+                 "[_locate_n] Unable to find an interval in $maxatt attempts";
+               return;
+            }
             $self->_nth_interval($last_int);
             my $ptr   = $$self{'data'}{'idate'}{$last_int}[1];
             if (defined $ptr) {
                my $date  = $$self{'data'}{'dates'}{$ptr};
-               last  if (defined $date  &&  $date->cmp($end) > 0);
+               if (defined($date)) {
+                  $n = 0;
+                  last  if ($date->cmp($end) > 0);
+               } else {
+                  $n++;
+               }
+            } else {
+               $n++;
             }
             $last_int++;
          }
@@ -1973,21 +2256,46 @@ sub _locate_n {
 
    $first_int    = $nn;
    if ($unmod) {
+      my $n = 0;
       while (1) {
+         if ($n > $maxatt) {
+            $$self{'err'} =
+              "[_locate_n] Unable to find an interval in $maxatt attempts";
+            return;
+         }
          $self->_nth_interval($first_int);
          my $date = $$self{'data'}{'idate'}{$first_int};
-         last  if (defined $date  &&  $date->cmp($start) < 0);
+         if (defined($date)) {
+            $n = 0;
+            last  if ($date->cmp($start) < 0);
+         } else {
+            $n++;
+         }
          $first_int--;
       }
 
    } else {
+      my $n = 0;
       LOOP:
       while (1) {
+         if ($n > $maxatt) {
+            $$self{'err'} =
+              "[_locate_n] Unable to find an interval in $maxatt attempts";
+            return;
+         }
          $self->_nth_interval($first_int);
          for (my $i=($first_int+1)*$ev - 1; $i >= $first_int*$ev; $i--) {
-            next  if (! exists $$self{'data'}{'dates'}{$i});
-            my $date = $$self{'data'}{'dates'}{$i};
-            last LOOP  if ($date->cmp($start) < 0);
+            if (exists $$self{'data'}{'dates'}{$i}) {
+               my $date = $$self{'data'}{'dates'}{$i};
+               if (defined($date)) {
+                  $n = 0;
+                  last LOOP  if ($date->cmp($start) < 0);
+               } else {
+                  $n++;
+               }
+            } else {
+               $n++;
+            }
          }
          $first_int--;
       }
@@ -1997,21 +2305,46 @@ sub _locate_n {
    #   i.e. Date(y) >= start  for modified dates
 
    if ($unmod) {
+      my $n = 0;
       while (1) {
+         if ($n > $maxatt) {
+            $$self{'err'} =
+              "[_locate_n] Unable to find an interval in $maxatt attempts";
+            return;
+         }
          $self->_nth_interval($first_int);
          my $date = $$self{'data'}{'idate'}{$first_int};
-         last  if (defined $date  &&  $date->cmp($start) >= 0);
+         if (defined($date)) {
+            $n = 0;
+            last  if ($date->cmp($start) >= 0);
+         } else {
+            $n++;
+         }
          $first_int++;
       }
 
    } else {
+      my $n = 0;
       LOOP:
       while (1) {
+         if ($n > $maxatt) {
+            $$self{'err'} =
+              "[_locate_n] Unable to find an interval in $maxatt attempts";
+            return;
+         }
          $self->_nth_interval($first_int);
          for (my $i=($first_int+1)*$ev - 1; $i >= $first_int*$ev; $i--) {
-            next  if (! exists $$self{'data'}{'dates'}{$i});
-            my $date = $$self{'data'}{'dates'}{$i};
-            last LOOP  if ($date->cmp($start) >= 0);
+            if (exists $$self{'data'}{'dates'}{$i}) {
+               my $date = $$self{'data'}{'dates'}{$i};
+               if (defined($date)) {
+                  $n = 0;
+                  last LOOP  if ($date->cmp($start) >= 0);
+               } else {
+                  $n++;
+               }
+            } else {
+               $n++;
+            }
          }
          $first_int++;
       }
@@ -2024,22 +2357,44 @@ sub _locate_n {
    $last_int = $first_int;
 
    if ($unmod) {
+      my $n = 0;
       while (1) {
+         if ($n > $maxatt) {
+            $$self{'err'} =
+              "[_locate_n] Unable to find an interval in $maxatt attempts";
+            return;
+         }
          $self->_nth_interval($last_int);
          my $date = $$self{'data'}{'idate'}{$last_int};
-         last  if (defined $date  &&  $date->cmp($end) > 0);
+         if (defined($date)) {
+            $n = 0;
+            last  if ($date->cmp($end) > 0);
+         } else {
+            $n++;
+         }
          $last_int++;
       }
       $last_int--;
 
    } else {
+      my $n = 0;
       LOOP:
       while (1) {
+         if ($n > $maxatt) {
+            $$self{'err'} =
+              "[_locate_n] Unable to find an interval in $maxatt attempts";
+            return;
+         }
          $self->_nth_interval($last_int);
          for (my $i=($last_int+1)*$ev - 1; $i >= $last_int*$ev; $i--) {
             next  if (! exists $$self{'data'}{'dates'}{$i});
             my $date = $$self{'data'}{'dates'}{$i};
-            last LOOP  if ($date->cmp($end) >= 0);
+            if (defined($date)) {
+               $n = 0;
+               last LOOP  if ($date->cmp($end) >= 0);
+            } else {
+               $n++;
+            }
          }
          $last_int++;
       }
@@ -2173,7 +2528,7 @@ sub _rtime_values {
    } elsif ($type eq 'dow_of_year') {
       my($y,$dow) = @args;
 
-      # Get the 1st occurence of $dow
+      # Get the 1st occurrence of $dow
       my $d0   = 1;
       my $dow0 = $dmb->day_of_week([$y,1,$d0]);
       if ($dow > $dow0) {
@@ -2201,7 +2556,7 @@ sub _rtime_values {
    } elsif ($type eq 'dow_of_month') {
       my($y,$m,$dow) = @args;
 
-      # Get the 1st occurence of $dow in the month
+      # Get the 1st occurrence of $dow in the month
       my $d0   = 1;
       my $dow0 = $dmb->day_of_week([$y,$m,$d0]);
       if ($dow > $dow0) {
